import os
import json

from .data import Data

__all__ = ['Jsonl']


class Jsonl:
	@classmethod
	def _traverse(cls, data, f):
		if isinstance(data, dict):
			return {key: cls._traverse(value, f) for key, value in data.items()}
		elif isinstance(data, list):
			return [cls._traverse(item, f) for item in data]
		elif isinstance(data, str):
			return f(data)
		else:
			return data

	@classmethod
	def _escape(cls, data):
		return cls._traverse(data, lambda s: s.replace('\n', '\\n'))
	
	@classmethod
	def _unescape(cls, data):
		return cls._traverse(data, lambda s: s.replace('\\n', '\n'))

	@classmethod
	def save(cls, data, location, max_size=99*1024):
		data = '\n'.join([cls.encode(item) for item in data])
		Data(data, 'jsonl').save(location, max_size)

	@classmethod
	def load(cls, location, generator=True, http_headers=None):
		data_object = Data.load(location, headers=http_headers)
		if generator:
			return (cls.decode(line) for line in data_object.lines)
		else:
			return [cls.decode(line) for line in data_object.lines]

	@classmethod
	def encode(cls, item):
		return json.dumps(cls._escape(item))

	@classmethod
	def decode(cls, line):
		return cls._unescape(json.loads(line.strip()))


import os
import csv

from .data import Data

__all__ = ['DataFrameRow', 'DataFrame']


class DataFrameRow(list):
	def __init__(self, *args, **kwargs):
		self.df = kwargs['df']
		super().__init__(*args)

	def set(self, col, value):
		col_idx = self.df.cols.get(col, None)
		if col_idx is not None:
			self[col_idx] = value

	def get(self, col):
		col_idx = self.df.cols.get(col, None)
		return self[col_idx] if col_idx is not None else ''

	def has(self, col):
		col_idx = self.df.cols.get(col, None)
		if col_idx:
			return len(self) > col_idx
		return False

	def remove(self, col):
		del self[self.df.cols[col]]

	def to_dict(self):
		return {self.df.header[n]:self[n] for n in range(len(self))}

	def copy(self, df):
		return DataFrameRow([*self], df=df)

class DataFrame:
	_id = 1
	
	######################################################

	@staticmethod
	def open(file_path, delimiter=None, encoding='utf-8'):
		df = DataFrame()
		df.append_file(file_path, delimiter, encoding)
		return df
		
	@staticmethod
	def load(location, http_headers=None):
		data_object   = Data.load(location, headers=http_headers)
		type_to_delim = {'csv': ',', 'tsv': '\t'}
		if data_object.file_type in type_to_delim:
			delimiter = type_to_delim[data_object.file_type]

		lines  = data_object.lines
		header = next(lines).split(delimiter)
		df = DataFrame(header)
		for line in lines:
			df.append(line.split(delimiter))
		return df

	######################################################
			
	def __init__(self, header=None, rows=None):
		self.id      = DataFrame._id
		self.rows    = []
		self.header  = None
		self.cols    = None
		self.indices = {}  # {col_name: {col_value.lower(): set(rowids)}}
		self.n       = 0

		if header:
			self.set_header(header)

		if rows:
			for row in rows:
				self.append(row)

		DataFrame._id += 1

	def __getitem__(self, key):  # [2:]
		return DataFrame(self.header, self.rows[key])

	def __len__(self):
		return len(self.rows)

	def __iter__(self):
		self.n = 0
		return self

	def __next__(self):
		if self.n < len(self):
			row = self.rows[self.n]
			self.n += 1
			return row
		else:
			raise StopIteration
	def _error(self, s):
		print('Error:', s)

	def get_index(self, col):
		return {
			row.get(col): row
			for row in self
		}

	def _update_header_positions(self):
		self.cols = { self.header[n]: n for n in range(len(self.header)) }

	def _join_headers(self, df):
		header = self.header.copy()
		for h in df.header:
			if h in header:
				header.append('_' + h)
			else:
				header.append(h)
		return header

	def set_header(self, header):
		self.header = header.copy()
		self._update_header_positions()
		return self

	def append(self, row):
		if not isinstance(row, DataFrameRow):
			row = DataFrameRow(row, df=self)
		row.df = self
		self.rows.append(row)
		return self

	def alter(self, cols, copy=False):
		df = DataFrame(cols)
		for row in self.rows:
			new_row = []
			for col in cols:
				new_row.append(row.get(col))
			df.append(new_row)
		if not copy:
			self.set_header(cols)
			self.rows = df.rows
			df = self
		return df

	def add_col(self, col, value='', after=None):
		position = len(self.header)
		if after in self.cols:
			position = self.cols[after] + 1
		self.header.insert(position, col)
		# Update columns positions
		self._update_header_positions()
		for row in self.rows:
			if callable(value):
				row.insert(position, value(row))
			else:
				row.insert(position, value)
		return self

	def add_cols(self, cols, after=None):
		for col in cols:
			self.add_col(col, after=after)
		return self

	def remove_col(self, col):
		col_idx = self.cols[col]
		del self.header[col_idx]
		for row in self.rows:
			del row[col_idx]
		self._update_header_positions()
		return self

	def remove_cols(self, cols):
		for col in cols:
			self.remove_col(col)
		return self

	def set(self, col, n, value):
		self.rows[n].set(col, value)
		return self

	def get(self, col, n):
		return self.rows[n].get(col)

	def map(self, f):  # f(row, n, df)
		for n in range(len(self.rows)):
			f(self.rows[n])
		return self

	def map_col(self, col, f):
		for n in range(len(self.rows)):
			self.rows[n].set(col, f(self.rows[n].get(col)))
		return self

	def map_cols(self, cols, f):
		for n in range(len(self.rows)):
			for col in cols:
				self.rows[n].set(col, f(self.rows[n].get(col)))
		return self

	def cast(self, col, _type=str):
		for row in self:
			row.set(col, _type(row.get(col)))
		return self

	def filter(self, f):
		df = DataFrame(self.header)
		for n in range(len(self.rows)):
			if f(self.rows[n], n):
				df.append(self.rows[n])
		return df

	def sort(self, f, cutoff=None):
		if cutoff is not None:
			items = sorted([(item, f(item)) for item in self.rows], key=lambda x: x[1])
			self.rows = [item for item, value in items if value > cutoff]
		else:
			self.rows.sort(key=f)

	def sort_by(self, col, asc=True):
		self.rows.sort(key=lambda row: row[self.cols[col]], reverse=not asc)

	def find(self, col, value):
		n = 0
		for row in self:
			if row.get(col) == value:
				return row, n
			n += 1
		return None, None

	def match(self, df, f):  # f(lrow, rrow) -> Falsey | new row
		for l in range(len(self.rows)):
			for r in range(len(df.rows)):
				f(self.rows[l], df.rows[r])
		return self

	def ljoin(self, df, f1, f2):
		header = self.header.copy() + [
			h if h not in self.header else f'_{h}' for h in df.header 
		]
		result_df = DataFrame(header)

		self.add_col('__join__1', f1)
		df.add_col('__join__2', f2)
		
		for row in self:
			value1    = row.get('__join__1')
			row2, _   = df.find('__join__2', value1)
			if row2 is not None:
				result_df.append(row[:-1] + row2[:-1])
			else:
				result_df.append(row[:-1] + [''] * (len(df.header) -1))

		self.remove_col('__join__1')
		df.remove_col('__join__2')
		return result_df

	def match_copy(self, df, header, f):  # f(lrow, rrow) -> Falsey | new row
		df1 = DataFrame(header)
		for l in self:
			for r in df:
				row = f(l, r)
				if row is not None:
					df1.append(row)
		return df1

	def transform(self, header, f):
		df = DataFrame(header)
		for row in self.rows:
			df.append(f(row))
		return df

	def append_file(self, file_path, delimiter=None, encoding='utf-8'):
		file_name, file_ext = os.path.splitext(file_path)
		if delimiter is None:
			delimiter = {'.tsv':'\t', '.csv':','}.get(file_ext.lower())

		with open(file_path, 'r', encoding=encoding) as f:
			reader = csv.reader(f, delimiter=delimiter)
			header = next(reader)
			if not self.header:
				self.set_header(header)
			for row in reader:
				self.append(row)
		return self

	def copy(self):
		df = DataFrame(self.header)
		for n in range(len(self.rows)):
			df.append(self.rows[n].copy(df))
		return df

	def save(self, location, file_type, max_size=99*1024):
		file_type = file_type.lower()
		delimiter = {'tsv':'\t', 'csv':','}.get(file_type)
		if delimiter:
			header = delimiter.join(self.header)
			rows   = [delimiter.join(row) for row in self]
			rows.insert(0, header)
			data = '\n'.join(rows)
			Data(data, file_type).save(location, max_size)
			return self
		else:
			raise Exception(f'Unsupported file type: "{file_type}"')


	def to_list(self, col, _type=str):
		return [_type(self.rows[n].get(col)) for n in range(len(self.rows))]

	def to_unique_list(self, col):
		values = set()
		for row in self:
			values.add(row.get(col))
		return list(values)

	def to_lists(self, cols):
		lists = []
		for col in cols:
			lists.append(self.to_list(col))
		return lists

	def print(self, max_rows=10):
		margin    = 3
		min_width = 12
		lens   = [(len(h) if len(h) > min_width else min_width) + margin for h in self.header]
		header = ''.join([self.header[n].ljust(lens[n]) for n in range(len(lens))])
		print(header)
		print('-' * len(header))
		count = 0
		for row in self.rows:
			if count < max_rows:
				s = ''.join([str(row[n])[:lens[n]-1].ljust(lens[n]) for n in range(len(lens))])
				print(s)
			else:
				break
			count += 1
		return self


	######################################################









import os
import json

from inspect import isgenerator
from pytest  import fixture
from puerml  import Jsonl


class TestJsonl:
	@fixture(autouse=True, scope='class', name='setup_TestJsonl')
	def setup(cls, request, tmp_path_factory):
		request.cls.location     = os.path.join(str(tmp_path_factory.mktemp('jsonl_test_dir')), 'test') 
		request.cls.web_location = 'https://raw.githubusercontent.com/PuerSoftware/puerml_test/main/jsonl_test_data'
		request.cls.test_data = [{
				'int'    : i,
				'string' : f'foo\n{i}',
				'list'   : [f'bar\n{j}' for j in range(i+1)],
				'dict'   : {f'baz{k}': k for k in range(i+1)}
			}
			for i in range(10)
		]

	def test_save_local(self):
		Jsonl.save(self.test_data, self.location)
	
	def test_load_local(self):
		loaded_data = Jsonl.load(self.location, generator=False)
		assert json.dumps(loaded_data) == json.dumps(self.test_data), 'Mismatch saved and loaded data'

	def test_load_local_generator(self):
		loaded_data = Jsonl.load(self.location)
		assert isgenerator(loaded_data)
		assert json.dumps(list(loaded_data)) == json.dumps(self.test_data), 'Mismatch saved and loaded data'

	def test_save_and_load_local_single_file(self):
		Jsonl.save(self.test_data, self.location, max_size=None)
		loaded_data = Jsonl.load(self.location, generator=False)
		assert json.dumps(loaded_data) == json.dumps(self.test_data), 'Mismatch saved and loaded data'

	def test_load_web(self):
		loaded_data = Jsonl.load(self.web_location, generator=False)
		assert json.dumps(loaded_data) == json.dumps(self.test_data), 'Mismatch saved and loaded data'
	


from functools           import wraps
from requests.exceptions import HTTPError, ConnectionError, Timeout, RequestException

def handle_http_error(f):
	@wraps(f)
	def _decorator(*args, **kwargs):
		try:
			return f(*args, **kwargs)
		except HTTPError as http_err:
			print(f'HTTP error occurred: {http_err}')
		except ConnectionError as conn_err:
			print(f'Connection error occurred: {conn_err}')
		except Timeout as timeout_err:
			print(f'Timeout error occurred: {timeout_err}')
		except RequestException as req_err:
			print(f'An error occurred: {req_err}')
		except Exception as e:
			print(f'An unexpected error occurred: {e}')	
	return _decorator
	


import os
import hashlib

from puerml  import Data
from pytest  import fixture


class TestData:
	@staticmethod
	def _make_txt_content(count):
		return {
			f'text_file_{i}.txt' : 'Hello world!\n' * (i+1)
			for i in range(count)
		}
	
	@staticmethod
	def _make_bin_content(count):
		return {
			f'bin_file_{i}' : b'Hello world!\n' * (i+1)
			for i in range(count)
		}

	@staticmethod
	def _write(file_name, content, mode):
		with open(file_name, mode) as f:
			f.write(content)
	
	@classmethod
	def _write_content(cls, location, content, mode):
		if not os.path.exists(location):
			os.makedirs(location)
		for file_name, c in content.items():
			cls._write(os.path.join(location, file_name), c, mode)

	@staticmethod
	def _read_checksum(file_path):
		with open(file_path, 'rb') as f:
			return hashlib.md5(f.read()).hexdigest()

	@classmethod
	def _get_checksums(cls, location):
		checksums = {}
		for root, _, files in os.walk(location):
			for file in files:
				file_path = os.path.join(root, file)
				checksums[file_path] = cls._read_checksum(file_path)
		return checksums
				
	def _assert_chesums(self, checksums):
		for file_path, checksum in self.checksums.items():
			assert checksums[file_path] == checksum

	@fixture(autouse=True, scope='class', name='setup_TestData')
	def setup(cls, request, tmp_path_factory):
		base_location  = str(tmp_path_factory.mktemp('data_test_dir'))
		location       = os.path.join(base_location, 'raw')
		zip_location   = os.path.join(base_location, 'zip')

		txt_content  = cls._make_txt_content(10)
		bin_content  = cls._make_bin_content(10)

		txt_location  = os.path.join(location, 'txt')
		bin_location  = os.path.join(location, 'bin')

		cls._write_content(txt_location, txt_content, 'w')
		cls._write_content(bin_location, bin_content, 'wb')
		
		checksums = cls._get_checksums(location)

		################################################
		  
		request.cls.location       = location
		request.cls.web_location   = 'https://raw.githubusercontent.com/PuerSoftware/puerml_test/main/data_test_data/zip'
		request.cls.zip_location   = zip_location
		request.cls.checksums      = checksums
		request.cls.max_chunk_size = 100

	def test_zip_local(self):
		data_object = Data(is_binary=True)
		data_object.zip(self.location)
		data_object.save(self.zip_location, self.max_chunk_size)

	def test_unzip_local(self):
		data_object = Data.load(self.zip_location, is_binary=True)
		data_object.unzip(self.location)
		checksums = self._get_checksums(self.location)
		self._assert_chesums(checksums)
	


from puerml  import DataFrame
from pytest  import fixture


class TestDataFrame:
	def _assert_df(self, df):
		for i in range(len(self.header)):
			col = f'h{i}'
			assert df.get(col, i) == self.df.get(col, i), '"Value mismatch in column {col} at index {i}'

	@fixture(autouse=True, scope='class', name='setup_TestDataFrame')
	def setup(cls, request, tmp_path_factory):
		request.cls.location       = str(tmp_path_factory.mktemp('df_test_dir'))
		request.cls.web_location   = 'https://raw.githubusercontent.com/PuerSoftware/puerml_test/main/df_test_data'
		request.cls.file_type      = 'tsv'
		request.cls.max_chunk_size = 1

		request.cls.header         = [f'h{i}' for i in range(10)]
		request.cls.rows           = [
			[f'row{i}-{j}' for j in range(10)]
			for i in range(10) 
		]
		request.cls.df = DataFrame(request.cls.header, request.cls.rows)

	def test_save_local(self):
		DataFrame(self.header, self.rows).save(self.location, self.file_type, self.max_chunk_size)
	
	def test_load_local(self):
		self._assert_df(DataFrame.load(self.location))
	
	def test_save_and_load_local_single_file(self):
		DataFrame(self.header, self.rows).save(self.location, self.file_type, None)
		self._assert_df(DataFrame.load(self.location))

	def test_load_web(self):
		self._assert_df(DataFrame.load(self.web_location))
		


import time
import json

__all__ = ['Benchmark']


class Benchmark:
    data = {}

    @staticmethod
    def time(func):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result     = func(*args, **kwargs)
            end_time   = time.time()

            # Update statistics
            func_name = func.__name__
            if func_name not in Benchmark.data:
                Benchmark.data[func_name] = [0, 0.0]  # Initialize [count, total_time]
            Benchmark.data[func_name][0] += 1  # Increment count
            Benchmark.data[func_name][1] += (end_time - start_time)  # Increment total time

            return result
        return wrapper

    @staticmethod
    def print():
        for f in Benchmark.data:
            print(f.ljust(30), str(Benchmark.data[f][0]).ljust(7), Benchmark.data[f][1])


import os
import shutil
import io
import json
import zipfile
import requests

__all__ = ['Data']


class Data:
	def __init__(self, data=None, file_type=None, is_binary=False):
		self.data       = io.BytesIO(data) if is_binary else io.StringIO(data)
		self.is_binary  = is_binary
		self.file_type  = file_type if data else None
		self.headers    = None
		self.location   = None
		self.from_web   = None
		self.description = ''

	def _read_manifest(self):
		file_loc = os.path.join(self.location, 'manifest.json')
		if self.from_web:
			response = requests.get(file_loc, headers=self.headers)
			return json.loads(response.text)
		else:
			with open(file_loc) as f:
				return json.load(f)

	def _write_manifest(self, location, manifest):
		file_loc = os.path.join(location, 'manifest.json')
		with open(file_loc, 'w') as f:
			return json.dump(manifest, f)

	def _read_web(self, file_loc):
		response = requests.get(file_loc, headers=self.headers, stream=True)
		if response.status_code == 200:
			return response
		return None

	def _read_disk(self, file_loc):
		try:
			if os.path.exists(file_loc):
				mode = 'rb' if self.is_binary else 'r'
				return open(file_loc, mode)
		except Exception as e:
			print(f'Failed to open file {file_loc}: {e}')
		return None

	def _file_generator(self):
		if self.data:
			yield self.data
		else:
			location = f'{self.location}.{self.file_type}'
			if os.path.exists(location):
				with open(location, 'rb' if self.is_binary else 'r') as f:
					yield f
			else:
				n = 0
				while True:
					file_name  = f'{n}.{self.file_type}' if self.file_type else str(n)
					file_loc = os.path.join(self.location, file_name)
					read = self._read_web if self.from_web else self._read_disk
					file = read(file_loc)
					if file:
						yield file
						if not self.from_web:
							file.close()
						n += 1
					else:
						break

	def _content_generator(self):  # TODO: modify to produce chunks efficiently
		for file in self._file_generator():
			if self.from_web:
				yield file.content if self.is_binary else file.text
			else:
				yield file.read()

	def _chunk_generator(self, max_size=1024*1024):
		for file in self._file_generator():
			if self.from_web:
				for chunk in file.iter_content(chunk_size=max_size):
					yield chunk if self.is_binary else chunk.decode('utf-8')
			else:
				while True:
					chunk = file.read(max_size)
					if not chunk:
						break
					yield chunk

	def _line_generator(self):
		for file in self._file_generator():
			iterator = file.iter_lines(decode_unicode=True) if self.from_web else file
			for line in iterator:
				yield line.strip()

	################################################################################

	@staticmethod
	def load(location, file_type=None, headers=None, is_binary=False):
		d = Data()
		d.data     = None
		d.location = location
		d.headers  = headers
		d.from_web = location.startswith('http')

		if file_type is None:
			manifest = d._read_manifest()
			d.description = manifest['description']
			d.is_binary   = manifest['is_binary']
			d.file_type   = manifest['type'].lower() if manifest['type'] else None
		else:
			d.description = ''
			d.is_binary   = is_binary
			d.file_type   = file_type

		return d

	def save(self, location, max_size=None, description=''):
		if max_size is None:
			location = f'{location}.{self.file_type}'
			with open(location, 'wb' if self.is_binary else 'w') as f:
				f.write(next(self._file_generator()).read())
			return

		if os.path.exists(location):
			shutil.rmtree(location)
		os.makedirs(location)

		def _write(l, t, n, m, c):
			fn = f'{n}.{t}' if t else str(n)
			l = os.path.join(l, fn)
			with open(l, m) as f:
				f.write(c)

		if self.is_binary:
			n = 0
			for chunk in self._chunk_generator(max_size):
				_write(location, self.file_type, n, 'wb', chunk)
				n += 1
		else:
			n = 0
			buffer, buffer_size = [], 0
			for line in self._line_generator():
				line_size        = len(line.encode('utf-8'))
				next_buffer_size = line_size + buffer_size
				if next_buffer_size > max_size and len(buffer) > 0:
					_write(location, self.file_type, n, 'w', '\n'.join(buffer))
					buffer, buffer_size = [], 0
					n += 1
				buffer.append(line)
				buffer_size = next_buffer_size
			if buffer_size > 0:
				_write(location, self.file_type, n, 'w', '\n'.join(buffer))
				n += 1

		self._write_manifest(location, {
			'description' : description or self.description,
			'count'       : n,
			'type'        : self.file_type,
			'is_binary'   : self.is_binary
		})

	def zip(self, location):
		self.data = io.BytesIO()
		with zipfile.ZipFile(self.data, 'w') as f:
			for root, _, files in os.walk(location):
				for file in files:
					file_path = os.path.join(root, file)
					f.write(file_path, os.path.relpath(file_path, start=location))
		self.data.seek(0)
		return self

	def unzip(self, location):
		buffer = io.BytesIO()
		for f in self._file_generator():
			buffer.write(f.read())

		buffer.seek(0)
		if os.path.exists(location):
			shutil.rmtree(location)
		with zipfile.ZipFile(buffer, 'r') as f:
			f.extractall(location)

	@property
	def lines(self):
		return self._line_generator()

	def content(self):
		content = ''
		for file_content in self._content_generator(): # TODO: add binary handling
			content += file_content
		return content

	def batches(self, batch_size=10): # TODO: add line preprocessing
		batch = []
		for line in self._line_generator():
			if len(batch) < batch_size:
				batch.append(line)
			else:
				yield batch
				batch = []
		if len(batch) > 0:
			yield batch

	def chunks(self, chunk_size=1024):
		...
